/*  cavemenu.go - a really dumb text menu system

LICENSE: BSD 3-CLAUSE

Copyright (c) 2024 hoff.industries

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

package cavemenu

import (
	"bufio"
	"fmt"
	"os"
	"regexp"
	"strconv"
	"strings"
)

type MenuErrorCode int64

const (
	NULL_APPEND        MenuErrorCode = 1
	NULL_MENU_REGISTER MenuErrorCode = 2
	CALLBACK_CONFLICT  MenuErrorCode = 3
	UNDERSIZED_QUEUE   MenuErrorCode = 4
	UNREGISTERED_MENU  MenuErrorCode = 5
	REGEX_MISMATCH     MenuErrorCode = 6
)

type MenuError struct {
	Code MenuErrorCode
}

func (e *MenuError) Error() string {
	return fmt.Sprintf("Menu Error: %d", e.Code)
}

type OptionMenuHandler struct {
	CallbackEvents chan int
	MenuRegister   map[int]*OptionMenu

	InputHandler *InputMenuHandler
	InputMenu    *InputMenu
}

type OptionMenu struct {
	Title string

	Submenus     []*OptionMenu
	CallbackCode int

	Marked bool
}

type InputMenuHandler struct {
	MenuRegister map[int]*InputMenu
}

type InputMenu struct {
	Title string
	Text  string
	Strip bool
	Retry bool

	RegexMatch   string
	CallbackCode int

	WordLimit int
}

func NewOptionHandler(queueSize int) (*OptionMenuHandler, error) {
	var newHandler = new(OptionMenuHandler)
	var menuError = new(MenuError)
	var registerError error

	if queueSize < 1 {
		menuError.Code = UNDERSIZED_QUEUE
		return nil, menuError
	}

	newHandler.CallbackEvents = make(chan int, queueSize)
	newHandler.MenuRegister = make(map[int]*OptionMenu)

	newHandler.InputHandler = new(InputMenuHandler)
	newHandler.InputHandler.MenuRegister = make(map[int]*InputMenu)

	newHandler.InputMenu = new(InputMenu)
	newHandler.InputMenu.Title = ""
	newHandler.InputMenu.Text = ""
	newHandler.InputMenu.Strip = true
	newHandler.InputMenu.Retry = true
	newHandler.InputMenu.CallbackCode = 0
	newHandler.InputMenu.RegexMatch = "^\\d+$"
	newHandler.InputMenu.WordLimit = 1

	registerError = newHandler.InputHandler.RegisterMenu(newHandler.InputMenu)

	if registerError != nil {
		return nil, registerError
	}

	return newHandler, nil
}

func NewInputHandler() *InputMenuHandler {
	var newHandler = new(InputMenuHandler)

	newHandler.MenuRegister = make(map[int]*InputMenu)
	return newHandler
}

func (menu *OptionMenu) MarkMenu(state bool) {
	menu.Marked = state
}

func (handler *OptionMenuHandler) NewMenu(title string,
	callbackCode int, submenus []*OptionMenu) (*OptionMenu, error) {

	var newMenu *OptionMenu = new(OptionMenu)
	var menuError *MenuError = new(MenuError)
	var mapLookup bool

	_, mapLookup = handler.MenuRegister[callbackCode]
	if mapLookup {
		menuError.Code = CALLBACK_CONFLICT
		return nil, menuError
	}

	newMenu.Marked = false
	newMenu.Submenus = submenus
	newMenu.Title = title
	newMenu.CallbackCode = callbackCode

	return newMenu, nil
}

func (handler *OptionMenuHandler) RegisterMenu(menu *OptionMenu) error {
	var menuError = new(MenuError)
	var mapLookup bool

	_, mapLookup = handler.MenuRegister[menu.CallbackCode]
	if mapLookup {
		menuError.Code = CALLBACK_CONFLICT
		return menuError
	}

	handler.MenuRegister[menu.CallbackCode] = menu
	return nil
}

func (handler *InputMenuHandler) RegisterMenu(menu *InputMenu) error {
	var menuError = new(MenuError)
	var mapLookup bool

	_, mapLookup = handler.MenuRegister[menu.CallbackCode]
	if mapLookup {
		menuError.Code = CALLBACK_CONFLICT
		return menuError
	}

	handler.MenuRegister[menu.CallbackCode] = menu
	return nil
}

func (menu *OptionMenu) AddSubmenu(submenu *OptionMenu) error {
	var menuError *MenuError = new(MenuError)

	if submenu == nil {
		menuError.Code = NULL_APPEND
		return menuError
	}

	if menu.Submenus == nil {
		menu.Submenus = make([]*OptionMenu, 1)
		menu.Submenus[0] = submenu
		return nil
	}

	menu.Submenus = append(menu.Submenus, submenu)
	return nil
}

func (menu *OptionMenu) AddSubmenus(submenus []*OptionMenu) error {
	var nilError = new(MenuError)
	var menuError error = nil
	nilError.Code = NULL_APPEND

	if submenus == nil {
		return menuError
	}

	for i := 0; i < len(submenus); i++ {
		menuError = menu.AddSubmenu(submenus[i])

		if menuError != nil {
			return menuError
		}
	}

	return nil
}

func (handler *InputMenuHandler) ExecuteMenu(callbackCode int, prompt string) (string, error) {
	var lineReader *bufio.Reader = bufio.NewReader(os.Stdin)

	var menu *InputMenu = nil
	var menuError = new(MenuError)

	var scanLine string = ""
	var scanError error

	var mapLookup bool
	var regexError error

	if handler.MenuRegister == nil {
		menuError.Code = NULL_MENU_REGISTER
		return "", menuError
	}

	menu, mapLookup = handler.MenuRegister[callbackCode]
	if !mapLookup {
		menuError.Code = UNREGISTERED_MENU
		return "", menuError
	}

	fmt.Printf("%s", menu.Title)
	fmt.Printf("%s", menu.Text)

	for {
		fmt.Printf("%s", prompt)

		scanLine, scanError = lineReader.ReadString('\n')

		if scanError != nil {
			fmt.Printf("Error reading input. Please try again.\n")
			continue
		}

		scanLine = strings.TrimSuffix(scanLine, "\n")
		if menu.Strip {
			scanLine = strings.TrimSuffix(scanLine, " ")
		}

		if scanLine == "" {
			continue
		}

		if menu.RegexMatch == "" {
			break
		}

		mapLookup, regexError = regexp.MatchString(menu.RegexMatch, scanLine)

		if regexError != nil || !mapLookup {

			if menu.Retry {
				fmt.Printf("Input did not match required format. Please try again.\n")
				continue
			}

			menuError.Code = REGEX_MISMATCH
			return "", menuError
		}

		break
	}

	return scanLine, nil
}

func (handler *OptionMenuHandler) ExecuteMenu(callbackCode int,
	prompt string, returnControl bool) (int, error) {

	var menuError = new(MenuError)
	var menu *OptionMenu = nil

	var scanError error
	var scanLine string
	var mapLookup bool

	var integerConversion int

	if handler.MenuRegister == nil {
		menuError.Code = NULL_MENU_REGISTER
		return -1, menuError
	}

	menu, mapLookup = handler.MenuRegister[callbackCode]
	if !mapLookup {
		menuError.Code = UNREGISTERED_MENU
		return -1, menuError
	}

	if menu.Submenus == nil {
		handler.CallbackEvents <- menu.CallbackCode
		return -1, nil
	}

	fmt.Printf("\n\t  ~~~ %s menu ~~~\n", menu.Title)

	for i := 0; i < len(menu.Submenus); i++ {
		fmt.Printf("\t")
		if menu.Submenus[i].Marked {
			fmt.Printf("* ")
		} else {
			fmt.Printf("  ")
		}
		fmt.Printf("%d) %s\n", i, menu.Submenus[i].Title)
	}

	fmt.Printf("\n")

	for {

		scanLine, scanError = handler.InputHandler.ExecuteMenu(0, prompt)
		if scanError != nil {
			return -1, scanError
		}

		integerConversion, scanError = strconv.Atoi(scanLine)

		if scanError != nil {
			fmt.Printf("Please input a number and try again.\n")
			continue
		}

		if integerConversion < 0 || integerConversion >= len(menu.Submenus) {
			fmt.Printf("Please select an option from the list and try again.\n")
			continue
		}

		break
	}

	// return the next menu in the sequence
	// so the caller can exec it after doing stuff
	if returnControl {
		handler.CallbackEvents <- menu.CallbackCode
		return menu.Submenus[integerConversion].CallbackCode, nil
	}

	return handler.ExecuteMenu(integerConversion, prompt, false)

}
