/*  gojp.go - CLI-based flashcard game for learning Japanese
    Copyright (C) 2024 hoff.industries

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>. */

package main

import (
	"encoding/csv"
	"encoding/json"
	"fmt"
	"github.com/rodaine/table"
	"hoff.industries/cavemenu"
	"hoff.industries/go.jp/colours"
	"hoff.industries/go.jp/definitions"
	"math/rand"
	"os"
	"path/filepath"
	"sort"
	"strconv"
	"strings"
	"time"
)

var AUTHOR string = "hoff.industries"
var VERSION string = "0.0.8"

type CharacterStatistics struct {
	character *gojpdefinitions.RomajiCharacter
	tries     int

	averageDuration time.Duration
	medianDuration  time.Duration
	modeDuration    time.Duration

	worstDuration time.Duration
	bestDuration  time.Duration

	gradient  int
	intercept int
}

type SessionResult struct {
	input    string
	result   bool
	duration time.Duration
}

type SessionStatistics struct {
	id        int
	startTime time.Time
	endTime   time.Time

	characterDuration time.Duration

	successRate int
	failureRate int
	total       int
	results     map[*gojpdefinitions.RomajiCharacter]*SessionResult
}

type ApplicationSettings struct {
	InfinitePractice bool
	PrioritiseWorst  bool

	Hiragana   bool
	Katakana   bool
	Handakuten bool
}

const (
	RET_SUCCESS int = 0
	RET_USER_INFO_ERR int = 1
	RET_FILE_OPEN_ERR int = 2
	RET_FILE_WRITE_ERR int = 3
	RET_GAME_EXEC_ERR int = 4
	RET_MENU_REGISTER_ERR int = 5
)

const (
	MENU_MAIN             int = 0x000
	MENU_PHONETICS        int = 0x002
	MENU_CHARACTERS_KANJI int = 0x008
	MENU_WORDS_HIRAGANA   int = 0x010
	MENU_WORDS_KATAKANA   int = 0x020
	MENU_WORDS_KANJI      int = 0x040

	MENU_STATISTICS int = 0x080
	MENU_SETTINGS   int = 0x100
	OPTION_EXIT     int = 0x800000

	OPTION_RANDOMISE_ALL int = 0x200
	MENU_BY_VOWEL        int = 0x400
	OPTION_FOR_ALL_CHARS int = 0x800

	OPTION_INFINITE_PRACTICE int = 0x1000
	OPTION_PRIORITISE_WORST  int = 0x2000

	OPTION_VOWEL_A int = 0x4000
	OPTION_VOWEL_E int = 0x8000
	OPTION_VOWEL_I int = 0x10000
	OPTION_VOWEL_O int = 0x20000
	OPTION_VOWEL_U int = 0x40000

	OPTION_BACK       int = 0x80000
	OPTION_HANDAKUTEN int = 0x100000
	OPTION_HIRAGANA   int = 0x200000
	OPTION_KATAKANA   int = 0x400000
)

var VOWELS_BY_OPTION = map[int][]string{
	OPTION_RANDOMISE_ALL: []string{"a", "e", "i", "o", "u"},

	OPTION_VOWEL_A: []string{"a"},
	OPTION_VOWEL_E: []string{"e"},
	OPTION_VOWEL_I: []string{"i"},
	OPTION_VOWEL_O: []string{"o"},
	OPTION_VOWEL_U: []string{"u"},
}

var MENU_DEFINITIONS = map[string][]*cavemenu.OptionMenu{
	"main": []*cavemenu.OptionMenu{
		&cavemenu.OptionMenu{Title: "main", Submenus: nil, CallbackCode: MENU_MAIN, Marked: false},
		&cavemenu.OptionMenu{Title: "back to main menu", Submenus: nil, CallbackCode: OPTION_BACK, Marked: true},
	},

	"main options": []*cavemenu.OptionMenu{

		&cavemenu.OptionMenu{Title: "practice phonetics", Submenus: nil, CallbackCode: MENU_PHONETICS, Marked: false},
		&cavemenu.OptionMenu{Title: "practice kanji", Submenus: nil, CallbackCode: MENU_CHARACTERS_KANJI, Marked: false},
		&cavemenu.OptionMenu{Title: "practice words - hiragana", Submenus: nil, CallbackCode: MENU_WORDS_HIRAGANA, Marked: false},
		&cavemenu.OptionMenu{Title: "practice words - katakana", Submenus: nil, CallbackCode: MENU_WORDS_KATAKANA, Marked: false},
		&cavemenu.OptionMenu{Title: "practice words - kanji", Submenus: nil, CallbackCode: MENU_WORDS_KANJI, Marked: false},

		&cavemenu.OptionMenu{Title: "statistics", Submenus: nil, CallbackCode: MENU_STATISTICS, Marked: false},
		&cavemenu.OptionMenu{Title: "settings", Submenus: nil, CallbackCode: MENU_SETTINGS, Marked: false},
		&cavemenu.OptionMenu{Title: "exit", Submenus: nil, CallbackCode: OPTION_EXIT, Marked: false},
	},

	"character specifications": []*cavemenu.OptionMenu{
		&cavemenu.OptionMenu{Title: "randomise all", Submenus: nil, CallbackCode: OPTION_RANDOMISE_ALL, Marked: false},
		&cavemenu.OptionMenu{Title: "by vowel", Submenus: nil, CallbackCode: MENU_BY_VOWEL, Marked: false},
		&cavemenu.OptionMenu{Title: "for all characters", Submenus: nil, CallbackCode: OPTION_FOR_ALL_CHARS, Marked: false},
		&cavemenu.OptionMenu{Title: "toggle hiragana", Submenus: nil, CallbackCode: OPTION_HIRAGANA, Marked: true},
		&cavemenu.OptionMenu{Title: "toggle katakana", Submenus: nil, CallbackCode: OPTION_KATAKANA, Marked: false},
		&cavemenu.OptionMenu{Title: "toggle han/dakuten", Submenus: nil, CallbackCode: OPTION_HANDAKUTEN, Marked: false},
	},

	"settings": []*cavemenu.OptionMenu{
		&cavemenu.OptionMenu{Title: "infinite practice mode", Submenus: nil, CallbackCode: OPTION_INFINITE_PRACTICE, Marked: false},
		&cavemenu.OptionMenu{Title: "prioritise worst characters", Submenus: nil, CallbackCode: OPTION_PRIORITISE_WORST, Marked: false},
	},

	"vowels": []*cavemenu.OptionMenu{
		&cavemenu.OptionMenu{Title: "a", Submenus: nil, CallbackCode: OPTION_VOWEL_A, Marked: false},
		&cavemenu.OptionMenu{Title: "e", Submenus: nil, CallbackCode: OPTION_VOWEL_E, Marked: false},
		&cavemenu.OptionMenu{Title: "i", Submenus: nil, CallbackCode: OPTION_VOWEL_I, Marked: false},
		&cavemenu.OptionMenu{Title: "o", Submenus: nil, CallbackCode: OPTION_VOWEL_O, Marked: false},
		&cavemenu.OptionMenu{Title: "u", Submenus: nil, CallbackCode: OPTION_VOWEL_U, Marked: false},
	},
}

func character_statistics_factory(characterStatistics *CharacterStatistics) {
	characterStatistics.character = nil
	characterStatistics.tries = 0

	characterStatistics.gradient = 0
	characterStatistics.intercept = 0
}

func clear_screen() {
	fmt.Print("\033[H\033[2J")
}

func generate_characterset(vowels []string,
	definitions map[string][]gojpdefinitions.RomajiCharacter, handakuten bool) []gojpdefinitions.RomajiCharacter {

	var characterSet []gojpdefinitions.RomajiCharacter = nil
	var vowelMap = make(map[string]bool)
	var mapLookup bool

	for i := 0; i < len(vowels); i++ {
		vowelMap[vowels[i]] = true
	}

	for i := range definitions {

		_, mapLookup = vowelMap[i]
		if !mapLookup {
			continue
		}

		for j := range definitions[i] {
			if !handakuten && definitions[i][j].HanDakuten {
				continue
			}
			characterSet = append(characterSet,
				definitions[i][j])
		}
	}

	return characterSet
}

func load_statistics(historyFile *os.File,
	sessionsFile *os.File) map[int]*SessionStatistics {

	var sessionStatistics *SessionStatistics = nil
	var sessionResult *SessionResult = nil
	var sessionMap = make(map[int]*SessionStatistics)
	var historyReader *csv.Reader

	var readRows [][]string
	var readRow []string
	var readError error
	var mapLookup bool

	var characterMap = make(map[string]*gojpdefinitions.RomajiCharacter)

	var character *gojpdefinitions.RomajiCharacter = nil

	var sessionId int

	if historyFile == nil || sessionsFile == nil {
		return nil
	}

	historyReader = csv.NewReader(sessionsFile)

	readRows, readError = historyReader.ReadAll()
	if readError != nil {
		return nil
	}

	for i := 1; i < len(readRows); i++ {
		readRow = readRows[i]

		if len(readRow) != 6 {
			return nil
		}

		sessionStatistics = new(SessionStatistics)

		sessionStatistics.id, readError = strconv.Atoi(readRow[0])
		if readError != nil {
			return nil
		}

		sessionStatistics.startTime, readError = time.Parse(
			"2006-01-02 15:04:05.999999999 -0700 MST", readRow[1])
		if readError != nil {
			return nil
		}

		sessionStatistics.successRate, readError = strconv.Atoi(readRow[3])
		if readError != nil {
			return nil
		}

		sessionStatistics.failureRate, readError = strconv.Atoi(readRow[4])
		if readError != nil {
			return nil
		}

		sessionStatistics.total, readError = strconv.Atoi(readRow[5])
		if readError != nil {
			return nil
		}

		if readRow[2] == "" {
			sessionMap[sessionStatistics.id] = sessionStatistics
			continue
		}

		sessionStatistics.endTime, readError = time.Parse(
			"2006-01-02 15:04:05.999999999 -0700 MST", readRow[2])
		if readError != nil {
			return nil
		}

		sessionMap[sessionStatistics.id] = sessionStatistics

	}

	historyReader = csv.NewReader(historyFile)

	readRows, readError = historyReader.ReadAll()
	if readError != nil {
		return nil
	}

	for i := 1; i < len(readRows); i++ {

		readRow = readRows[i]

		if len(readRow) != 5 {
			return nil
		}

		sessionId, readError = strconv.Atoi(readRow[0])
		if readError != nil {
			return nil
		}

		sessionStatistics, mapLookup = sessionMap[sessionId]
		if !mapLookup {
			return nil
		}

		if sessionStatistics.results == nil {
			sessionStatistics.results = make(
				map[*gojpdefinitions.RomajiCharacter]*SessionResult)
		}

		sessionResult = new(SessionResult)

		sessionResult.input = readRow[2]
		sessionResult.result = readRow[3] == "true"

		sessionResult.duration, readError = time.ParseDuration(readRow[4])
		if readError != nil {
			return nil
		}

		sessionStatistics.characterDuration = sessionStatistics.characterDuration + sessionResult.duration

		character, mapLookup = characterMap[readRow[1]]
		if !mapLookup {
			characterMap[readRow[1]] = new(gojpdefinitions.RomajiCharacter)
			character = characterMap[readRow[1]]
			character.Character = readRow[1]

		}

		sessionStatistics.results[character] = sessionResult
	}

	return sessionMap
}

func write_character(sessionId int, character *gojpdefinitions.RomajiCharacter,
	sessionResult *SessionResult, historyFile *os.File) bool {
	var writeSlice = []string{
		fmt.Sprintf("%d", sessionId),
		character.Character,
		sessionResult.input,
		strconv.FormatBool(sessionResult.result),
		sessionResult.duration.String(),
	}

	return write_slice(writeSlice, historyFile)
}

func write_session(sessionStatistics SessionStatistics, sessionsFile *os.File) bool {

	var writeSlice = []string{
		fmt.Sprintf("%d", sessionStatistics.id),
		sessionStatistics.startTime.Format("2006-01-02 15:04:05.999999999 -0700 MST"),
		"",
		fmt.Sprintf("%d", sessionStatistics.successRate),
		fmt.Sprintf("%d", sessionStatistics.failureRate),
		fmt.Sprintf("%d", sessionStatistics.total),
	}

	if !(sessionStatistics.endTime.IsZero()) {
		writeSlice[2] = sessionStatistics.endTime.Format("2006-01-02 15:04:05.999999999 -0700 MST")
	}

	return write_slice(writeSlice, sessionsFile)
}

func write_slice(writeSlice []string, file *os.File) bool {

	var fileWriter *csv.Writer

	if writeSlice == nil || file == nil {
		return false
	}

	fileWriter = csv.NewWriter(file)
	if (fileWriter.Write(writeSlice) != nil) {
		return false
	}
	fileWriter.Flush()
	return true
}

func write_history(sessionStatistics SessionStatistics,
	historyFile *os.File, sessionsFile *os.File) bool {

	var sessionResult *SessionResult

	write_session(sessionStatistics, sessionsFile)

	for i := range sessionStatistics.results {
		sessionResult = sessionStatistics.results[i]

		write_character(sessionStatistics.id, i, sessionResult, historyFile)

	}

	return true
}

func calculate_totals(statistics map[int]*SessionStatistics,
	vowelList []string) map[*gojpdefinitions.RomajiCharacter]*CharacterStatistics {

	var totals = make(map[*gojpdefinitions.RomajiCharacter]*CharacterStatistics)
	var sessionStatistics *SessionStatistics = nil
	var characterStatistics *CharacterStatistics = nil

	var durations = make(map[*gojpdefinitions.RomajiCharacter][]int64)

	var sessionResult *SessionResult = nil

	var vowelMap = make(map[string]bool)
	var mapLookup bool

	for i := 0; i < len(vowelList); i++ {
		vowelMap[vowelList[i]] = true
	}

	for i := range statistics {
		sessionStatistics = statistics[i]

		for j := range sessionStatistics.results {

			/*_, mapLookup = vowelMap[j.Family]
			if (!mapLookup) {
				continue
			}*/

			sessionResult = sessionStatistics.results[j]

			characterStatistics, mapLookup = totals[j]
			if !mapLookup {
				totals[j] = new(CharacterStatistics)
				character_statistics_factory(totals[j])

				characterStatistics = totals[j]
			}

			_, mapLookup = durations[j]
			if !mapLookup {
				durations[j] = make([]int64, 1)
				durations[j][0] = (int64)(sessionResult.duration)
			} else {
				durations[j] = append(durations[j], (int64)(sessionResult.duration))
			}

			if (int64)(characterStatistics.bestDuration) == 0 {
				characterStatistics.bestDuration = sessionResult.duration
			}

			characterStatistics.tries++
			characterStatistics.averageDuration = characterStatistics.averageDuration + sessionResult.duration

			if sessionResult.duration < characterStatistics.bestDuration {
				characterStatistics.bestDuration = sessionResult.duration
			}

			if sessionResult.duration > characterStatistics.worstDuration {
				characterStatistics.worstDuration = sessionResult.duration
			}

		}

	}

	for i := range totals {
		characterStatistics = totals[i]
		characterStatistics.averageDuration = (time.Duration)(
			((int64)(characterStatistics.averageDuration)) / (int64)(characterStatistics.tries))
	}

	return totals
}

func display_statistics(sessionStatistics SessionStatistics,
	characterSet []gojpdefinitions.RomajiCharacter) {

	var sessionResult *SessionResult

	var charactersOrdered []*gojpdefinitions.RomajiCharacter

	fmt.Printf("Session Statistics:\n")

	fmt.Printf("%s%d%s / %d\n", gojpcolours.RESULT_FORMAT[true],
		sessionStatistics.successRate, gojpcolours.COLOURS["CLEAR"], len(characterSet))
	fmt.Printf("%s%d%s / %d\n", gojpcolours.RESULT_FORMAT[false],
		sessionStatistics.failureRate, gojpcolours.COLOURS["CLEAR"], len(characterSet))

	fmt.Printf("Completion Time: %s\n",
		sessionStatistics.characterDuration.String())

	for i := range sessionStatistics.results {
		charactersOrdered = append(charactersOrdered, i)
	}

	sort.Slice(charactersOrdered, func(i, j int) bool {
		return sessionStatistics.results[charactersOrdered[i]].duration < sessionStatistics.results[charactersOrdered[j]].duration
	})

	for i := 0; i < len(charactersOrdered); i++ {
		sessionResult = sessionStatistics.results[charactersOrdered[i]]

		fmt.Printf("\t%s%s - %s - %s\n", gojpcolours.RESULT_FORMAT[sessionResult.result],
			charactersOrdered[i].Character, sessionResult.input, sessionResult.duration.String())
	}

	sessionResult = sessionStatistics.results[charactersOrdered[0]]

	fmt.Printf("The fastest character for you to recognise was %s at %s.\n",
		charactersOrdered[0].Character, sessionResult.duration.String())

	sessionResult = sessionStatistics.results[charactersOrdered[len(charactersOrdered)-1]]

	fmt.Printf("The longest character for you to recognise was %s at %s.\n",
		charactersOrdered[len(charactersOrdered)-1].Character, sessionResult.duration.String())
}

func game_loop(sessionId int, characterSet []gojpdefinitions.RomajiCharacter,
	historyFile *os.File, sessionsFile *os.File) *SessionStatistics {

	var scanError error
	var scanLine string = ""
	var scanResult bool

	var inputHandler = cavemenu.NewInputHandler()
	var inputMenu = new(cavemenu.InputMenu)

	var sessionStatistics = new(SessionStatistics)
	sessionStatistics.results = make(
		map[*gojpdefinitions.RomajiCharacter]*SessionResult)

	var romajiCharacter gojpdefinitions.RomajiCharacter
	var sessionResult *SessionResult

	var characterStartTime time.Time

	if characterSet == nil {
		return nil
	}

	inputMenu.CallbackCode = 0
	inputMenu.Strip = true
	inputMenu.Retry = false
	inputMenu.WordLimit = 1

	if (inputHandler.RegisterMenu(inputMenu) != nil) {
		return nil
	}

	rand.Shuffle(len(characterSet), func(i, j int) {
		characterSet[i], characterSet[j] = characterSet[j], characterSet[i]
	})

	clear_screen()

	fmt.Printf("\n0 / %d\n0 / %d\n", len(characterSet), len(characterSet))

	for j := 5; j > 0; j-- {
		fmt.Printf(" Starting in %d seconds...\r", j)
		time.Sleep(time.Second)
	}

	clear_screen()
	fmt.Printf("\n")

	sessionStatistics.total = len(characterSet)
	sessionStatistics.startTime = time.Now()

	for i := 0; i < sessionStatistics.total; i++ {

		sessionResult = new(SessionResult)
		romajiCharacter = characterSet[i]

		characterStartTime = time.Now()

		inputMenu.Title = fmt.Sprintf("%s%d%s / %d\n%s%d%s / %d\n", gojpcolours.RESULT_FORMAT[true],
			sessionStatistics.successRate, gojpcolours.COLOURS["CLEAR"],
			len(characterSet), gojpcolours.RESULT_FORMAT[false],
			sessionStatistics.failureRate, gojpcolours.COLOURS["CLEAR"], len(characterSet))

		inputMenu.Text = fmt.Sprintf("\t** %s **\n", characterSet[i].Character)

		inputMenu.RegexMatch = "^("
		for j := 0; j < len(romajiCharacter.ValidRomaji); j++ {
			inputMenu.RegexMatch += fmt.Sprintf("%s|", romajiCharacter.ValidRomaji[j])
		}
		inputMenu.RegexMatch = strings.TrimSuffix(inputMenu.RegexMatch, "|")
		inputMenu.RegexMatch = inputMenu.RegexMatch + ")+$"

		_, scanError = inputHandler.ExecuteMenu(0, gojpcolours.PS1)
		scanResult = scanError == nil

		sessionStatistics.endTime = time.Now()

		sessionResult.result = scanResult
		sessionResult.input = scanLine
		sessionResult.duration = sessionStatistics.endTime.Sub(characterStartTime)

		sessionStatistics.results[gojpdefinitions.Romaji2Heap(romajiCharacter)] = sessionResult

		clear_screen()

		if scanResult {

			fmt.Printf("\t%sはい!%s\n",
				gojpcolours.RESULT_FORMAT[true], gojpcolours.COLOURS["CLEAR"])

			sessionStatistics.successRate++
			continue
		}

		fmt.Printf("\t%sいいえ...%s\nValid romaji for ** %s ** :\n",
			gojpcolours.RESULT_FORMAT[false],
			gojpcolours.COLOURS["CLEAR"], romajiCharacter.Character)

		for j := 0; j < len(romajiCharacter.ValidRomaji); j++ {
			fmt.Printf("\t%s\n", romajiCharacter.ValidRomaji[j])
		}

		for j := 5; j > 0; j-- {
			fmt.Printf(" Try another one in %d seconds...\r", j)
			time.Sleep(time.Second)
		}

		clear_screen()
		fmt.Printf("\n")
		sessionStatistics.failureRate++

	}

	sessionStatistics.endTime = time.Now()

	for i := range sessionStatistics.results {
		sessionResult = sessionStatistics.results[i]

		sessionStatistics.characterDuration += sessionResult.duration
	}

	sessionStatistics.id = sessionId
	if !write_history(*sessionStatistics, historyFile, sessionsFile) {
		fmt.Printf("Warning! Couldn't write statistics to disk!\n")
		return nil
	}

	return sessionStatistics
}

func display_totals(totals map[*gojpdefinitions.RomajiCharacter]*CharacterStatistics) {
	var characterStatistics *CharacterStatistics = nil
	var totalsOrdered []*gojpdefinitions.RomajiCharacter

	var totalsTable table.Table
	var tableInterface []interface{} = make([]interface{}, 4)
	var headers = []string{
		"character",
		"avg. time",
		"best time",
		"worst time",
	}

	for i := range totals {
		totalsOrdered = append(totalsOrdered, i)
	}

	for i := 0; i < 4; i++ {
		tableInterface[i] = headers[i]
	}

	totalsTable = table.New(tableInterface...)

	sort.Slice(totalsOrdered, func(i, j int) bool {
		return totals[totalsOrdered[i]].averageDuration < totals[totalsOrdered[j]].averageDuration
	})

	for i := 0; i < len(totalsOrdered); i++ {

		tableInterface = make([]interface{}, 4)

		characterStatistics = totals[totalsOrdered[i]]

		tableInterface[0] = fmt.Sprintf("%s**%s %s %s**%s",
			gojpcolours.COLOURS["RED_BOLD"], gojpcolours.COLOURS["SAKURA_BOLD"],
			totalsOrdered[i].Character, gojpcolours.COLOURS["RED_BOLD"], gojpcolours.COLOURS["CLEAR"])

		tableInterface[1] = fmt.Sprintf("%savg. time:%s %s%s",
			gojpcolours.COLOURS["BLUE"], gojpcolours.COLOURS["WHITE"],
			characterStatistics.averageDuration.String(), gojpcolours.COLOURS["CLEAR"])

		tableInterface[2] = fmt.Sprintf("%sbest time:%s %s%s",
			gojpcolours.COLOURS["GREEN"], gojpcolours.COLOURS["WHITE"],
			characterStatistics.bestDuration.String(), gojpcolours.COLOURS["CLEAR"])

		tableInterface[3] = fmt.Sprintf("%sworst time:%s %s%s",
			gojpcolours.COLOURS["RED_BOLD"], gojpcolours.COLOURS["WHITE"],
			characterStatistics.worstDuration.String(), gojpcolours.COLOURS["CLEAR"])

		totalsTable.AddRow(tableInterface...)
	}

	totalsTable.Print()

	fmt.Printf("Your best character is %s with at an average of %s.\n",
		totalsOrdered[0].Character, totals[totalsOrdered[0]].averageDuration)

	fmt.Printf("Your worst character is %s with an average of %s.\n",
		totalsOrdered[len(totalsOrdered)-1].Character,
		totals[totalsOrdered[len(totalsOrdered)-1]].averageDuration)
}

func write_settings(applicationSettings ApplicationSettings, settingsFile *os.File) bool {

	var settingsJson []byte
	var jsonError error

	if settingsFile == nil {
		return false
	}

	jsonError = settingsFile.Truncate(0)
	if jsonError != nil {
		return false
	}
	_, jsonError = settingsFile.Seek(0, 0)
	if jsonError != nil {
		return false
	}

	settingsJson, jsonError = json.Marshal(applicationSettings)
	if jsonError != nil {
		return false
	}

	_, jsonError = settingsFile.Write(settingsJson)

	if (jsonError != nil) {
		return false
	}

	return settingsFile.Sync() == nil
}

func load_settings(applicationSettings *ApplicationSettings, settingsPath string) bool {

	var jsonError error
	var settingsContents []byte

	settingsContents, jsonError = os.ReadFile(settingsPath)
	if jsonError != nil {
		return false
	}

	jsonError = json.Unmarshal(settingsContents, applicationSettings)
	return jsonError == nil
}

func main() {

	var menuHandler *cavemenu.OptionMenuHandler
	var menuMask int = 0
	var nextMenu int

	var configDirectory string
	var filePath string
	var userError error
	var mapLookup bool
	var sessionId int = -1

	var historyFile *os.File
	var sessionsFile *os.File
	var settingsFile *os.File

	var lastSession *SessionStatistics = nil

	var userStatistics map[int]*SessionStatistics
	var callbackEvent int = -1

	var applicationSettings = &ApplicationSettings{
		InfinitePractice: false,
		PrioritiseWorst:  false,
		Hiragana:         true,
		Katakana:         false,
		Handakuten:       false,
	}

	var breakFlag bool = false
	var previousMenu int

	var vowelOptions int = OPTION_VOWEL_A | OPTION_VOWEL_E | OPTION_VOWEL_I |
		OPTION_VOWEL_O | OPTION_VOWEL_U | OPTION_RANDOMISE_ALL

	var settingsMask int = OPTION_HANDAKUTEN | OPTION_HIRAGANA |
		OPTION_KATAKANA | OPTION_INFINITE_PRACTICE | OPTION_PRIORITISE_WORST

	var characterSet []gojpdefinitions.RomajiCharacter = nil
	var sessionStatistics *SessionStatistics

	menuHandler, _ = cavemenu.NewOptionHandler(256)

	rand.Seed(time.Now().UnixNano())

	fmt.Printf("いらっしゃいませ \n%s\n\nversion %s by %s\n", gojpcolours.ASCII, VERSION, AUTHOR)

	configDirectory, userError = os.UserHomeDir()
	if userError != nil {
		fmt.Printf("Couldn't obtain your home directory! Exiting...\n")
		os.Exit(RET_USER_INFO_ERR)
	}

	configDirectory = filepath.Join(configDirectory, ".config")
	configDirectory = filepath.Join(configDirectory, "gojp")

	userError = os.Mkdir(configDirectory, 0755)

	filePath = filepath.Join(configDirectory, "history.csv")

	_, userError = os.Stat(filePath)
	if userError != nil {
		historyFile, userError = os.Create(filePath)

		historyWriter := csv.NewWriter(historyFile)
		if (historyWriter.Write([]string{"session_id", "character",
			"input", "result", "duration"}) != nil) {
				os.Exit(RET_FILE_WRITE_ERR)
			}
		historyWriter.Flush()

	} else {
		historyFile, userError = os.OpenFile(
			filePath, os.O_RDONLY|os.O_CREATE|os.O_APPEND, 0777)
	}

	if userError != nil {
		fmt.Printf("Couldn't open history file! History will not be loaded.\n")
		historyFile = nil
	}

	filePath = filepath.Join(configDirectory, "sessions.csv")

	_, userError = os.Stat(filePath)
	if userError != nil {
		sessionsFile, userError = os.Create(filePath)

		historyWriter := csv.NewWriter(sessionsFile)
		if (historyWriter.Write([]string{"session_id", "session_start",
			"session_end", "success", "failure", "total"}) != nil) {
				os.Exit(RET_FILE_WRITE_ERR)
			}
		historyWriter.Flush()

	} else {
		sessionsFile, userError = os.OpenFile(
			filePath, os.O_RDONLY|os.O_CREATE|os.O_APPEND, 0777)
	}

	if userError != nil {
		fmt.Printf("Couldn't open session file! Historical statistics will not be loaded.\n")
		sessionsFile = nil
	}

	filePath = filepath.Join(configDirectory, "settings.json")

	load_settings(applicationSettings, filePath)

	settingsFile, userError = os.OpenFile(
		filePath, os.O_RDWR|os.O_CREATE, 0755)

	if userError != nil {
		fmt.Printf("Couldn't open session file! Historical statistics will not be loaded.\n")
		sessionsFile = nil
	}

	userStatistics = load_statistics(historyFile, sessionsFile)
	if userStatistics == nil {
		fmt.Printf("Couldn't load statistics, format invalid! Historical statistics will not be recorded.\n")

		sessionsFile.Close()
		historyFile.Close()

		sessionsFile = nil
		historyFile = nil
	} else {

		for i := range userStatistics {
			if i > sessionId {
				sessionId = i
			}
		}

		lastSession, mapLookup = userStatistics[sessionId]
		if !mapLookup {
			lastSession = nil
		}

		sessionId++

		sessionsFile.Close()
		historyFile.Close()

		filePath = filepath.Join(configDirectory, "sessions.csv")

		sessionsFile, userError = os.OpenFile(
			filePath, os.O_WRONLY|os.O_CREATE|os.O_APPEND, 0644)
		if (userError != nil) {
			os.Exit(RET_FILE_OPEN_ERR)
		}

		filePath = filepath.Join(configDirectory, "history.csv")

		historyFile, userError = os.OpenFile(
			filePath, os.O_WRONLY|os.O_CREATE|os.O_APPEND, 0644)
		if (userError != nil) {
			os.Exit(RET_FILE_OPEN_ERR)
		}

		if lastSession != nil {

			if lastSession.endTime.IsZero() {
				fmt.Printf("\nWelcome back. Your last session was %sincomplete%s, but so far you had scored %s%d %s/ %d.",
					gojpcolours.COLOURS["ORANGE_BOLD"], gojpcolours.COLOURS["CLEAR"], gojpcolours.COLOURS["ORANGE_BOLD"], lastSession.successRate, gojpcolours.COLOURS["CLEAR"], lastSession.total)
			} else {
				fmt.Printf("\nYour last session was at %s.\nYou scored %s%d %s/ %d, and it took you %s.\n",
					lastSession.startTime, gojpcolours.RESULT_FORMAT[lastSession.successRate > (lastSession.total/2)],
					lastSession.successRate, gojpcolours.COLOURS["CLEAR"], lastSession.total, lastSession.characterDuration.String())
			}

		} else {
			fmt.Printf("\nThanks for using go.jp! Just have fun with it!\n")
		}

	}

	for i := range MENU_DEFINITIONS {

		for j := 0; j < len(MENU_DEFINITIONS[i]); j++ {
			if (menuHandler.RegisterMenu(MENU_DEFINITIONS[i][j]) != nil) {
				os.Exit(RET_MENU_REGISTER_ERR)
			}
		}
	}

	menuHandler.MenuRegister[MENU_MAIN].AddSubmenus(MENU_DEFINITIONS["main options"])
	menuHandler.MenuRegister[MENU_SETTINGS].AddSubmenus(MENU_DEFINITIONS["settings"])
	menuHandler.MenuRegister[MENU_SETTINGS].AddSubmenu(menuHandler.MenuRegister[OPTION_BACK])

	menuHandler.MenuRegister[MENU_PHONETICS].AddSubmenu(menuHandler.MenuRegister[OPTION_HIRAGANA])
	menuHandler.MenuRegister[MENU_PHONETICS].AddSubmenu(menuHandler.MenuRegister[OPTION_KATAKANA])

	for i := MENU_PHONETICS; i < MENU_CHARACTERS_KANJI; i = MENU_CHARACTERS_KANJI {
		menuHandler.MenuRegister[i].AddSubmenu(menuHandler.MenuRegister[OPTION_HANDAKUTEN])
		menuHandler.MenuRegister[i].AddSubmenu(menuHandler.MenuRegister[OPTION_RANDOMISE_ALL])
		menuHandler.MenuRegister[i].AddSubmenu(menuHandler.MenuRegister[MENU_BY_VOWEL])
		menuHandler.MenuRegister[i].AddSubmenu(menuHandler.MenuRegister[OPTION_BACK])
	}

	menuHandler.MenuRegister[MENU_STATISTICS].AddSubmenu(menuHandler.MenuRegister[OPTION_FOR_ALL_CHARS])
	menuHandler.MenuRegister[MENU_STATISTICS].AddSubmenu(menuHandler.MenuRegister[MENU_BY_VOWEL])
	menuHandler.MenuRegister[MENU_STATISTICS].AddSubmenu(menuHandler.MenuRegister[OPTION_BACK])
	menuHandler.MenuRegister[MENU_BY_VOWEL].AddSubmenus(MENU_DEFINITIONS["vowels"])

	menuHandler.MenuRegister[OPTION_KATAKANA].MarkMenu(applicationSettings.Katakana)
	menuHandler.MenuRegister[OPTION_HIRAGANA].MarkMenu(applicationSettings.Hiragana)
	menuHandler.MenuRegister[OPTION_HANDAKUTEN].MarkMenu(applicationSettings.Handakuten)
	menuHandler.MenuRegister[OPTION_INFINITE_PRACTICE].MarkMenu(applicationSettings.InfinitePractice)
	menuHandler.MenuRegister[OPTION_PRIORITISE_WORST].MarkMenu(applicationSettings.PrioritiseWorst)

	nextMenu, _ = menuHandler.ExecuteMenu(0, gojpcolours.PS1, true)

	for {

		previousMenu = callbackEvent

		callbackEvent = <-menuHandler.CallbackEvents
		menuMask = menuMask | callbackEvent

		if (OPTION_BACK & menuMask) == OPTION_BACK {
			nextMenu = MENU_MAIN
			menuMask = 0x00
		}

		switch menuMask {

		case MENU_STATISTICS | OPTION_FOR_ALL_CHARS:
			display_totals(calculate_totals(userStatistics,
				[]string{"a", "e", "i", "o", "u"}))

		case MENU_SETTINGS | OPTION_INFINITE_PRACTICE:
			applicationSettings.InfinitePractice = !(applicationSettings.InfinitePractice)
			menuHandler.MenuRegister[OPTION_INFINITE_PRACTICE].MarkMenu(applicationSettings.InfinitePractice)

			nextMenu = previousMenu

		case MENU_SETTINGS | OPTION_PRIORITISE_WORST:
			applicationSettings.PrioritiseWorst = !(applicationSettings.PrioritiseWorst)
			menuHandler.MenuRegister[OPTION_PRIORITISE_WORST].MarkMenu(applicationSettings.PrioritiseWorst)

			nextMenu = previousMenu

		case MENU_PHONETICS | OPTION_HIRAGANA:
			applicationSettings.Hiragana = !(applicationSettings.Hiragana)
			menuHandler.MenuRegister[OPTION_HIRAGANA].MarkMenu(applicationSettings.Hiragana)

			nextMenu = previousMenu

		case MENU_PHONETICS | OPTION_KATAKANA:
			applicationSettings.Katakana = !(applicationSettings.Katakana)
			menuHandler.MenuRegister[OPTION_KATAKANA].MarkMenu(applicationSettings.Katakana)

			nextMenu = previousMenu

		case MENU_PHONETICS | OPTION_HANDAKUTEN:
			applicationSettings.Handakuten = !(applicationSettings.Handakuten)
			menuHandler.MenuRegister[OPTION_HANDAKUTEN].MarkMenu(applicationSettings.Handakuten)

			nextMenu = previousMenu

		case OPTION_EXIT:
			breakFlag = true
		}

		if breakFlag {
			break
		}

		if (menuMask & settingsMask) != 0x00 {
			write_settings(*applicationSettings, settingsFile)
			menuMask = menuMask &^ settingsMask
		}

		if (menuMask&MENU_PHONETICS) == MENU_PHONETICS &&
			(menuMask&vowelOptions != 0x00) {

			if applicationSettings.Hiragana {
				characterSet = append(characterSet, generate_characterset(
					VOWELS_BY_OPTION[menuMask & ^MENU_PHONETICS & ^MENU_BY_VOWEL],
					gojpdefinitions.ROMAJI_HIRAGANA, applicationSettings.Handakuten)...)
			}

			if applicationSettings.Katakana {
				characterSet = append(characterSet, generate_characterset(
					VOWELS_BY_OPTION[menuMask & ^MENU_PHONETICS & ^MENU_BY_VOWEL],
					gojpdefinitions.ROMAJI_KATAKANA, applicationSettings.Handakuten)...)
			}

			sessionStatistics = game_loop(sessionId, characterSet, historyFile, sessionsFile)
			if sessionStatistics == nil {
				os.Exit(RET_GAME_EXEC_ERR)
			}

			display_statistics(*sessionStatistics, characterSet)
			fmt.Printf("Session statistics written to disk.%s\n",
				gojpcolours.COLOURS["CLEAR"])
			os.Exit(RET_SUCCESS)

		}

		if nextMenu == -1 {
			nextMenu = MENU_MAIN
			menuMask = 0x00
		}

		nextMenu, _ = menuHandler.ExecuteMenu(nextMenu, gojpcolours.PS1, true)

	}

	if historyFile != nil {
		historyFile.Close()
	}

	if sessionsFile != nil {
		sessionsFile.Close()
	}

	os.Exit(RET_SUCCESS)
}
